#include <vector>
#include <cstdio>
#include <cstdlib>
#include <sys/time.h>
#include <float.h>
#include <ANN/ANN.h>

#include "CUDA_NN.h"

using namespace std;

double TimeDiff(timeval t1, timeval t2);
double SearchCPU(const vector <float> &query, const vector <float> &data, vector <int> &idxs, vector <float> dist_sq);
double SearchANN(const vector <float> &query, const vector <float> &data, vector <int> &idxs, vector <float> dist_sq);

int main()
{
    timeval t1, t2;
    int data_pts = 100000;
    int query_pts = 100000;

    vector <float> data(data_pts*CUDA_NN_DIM);
    vector <float> query(query_pts*CUDA_NN_DIM);
    vector <int> idxs(query_pts);
    vector <float> dist_sq(query_pts);
    vector <int> cpu_idxs(query_pts);
    vector <float> cpu_dist_sq(query_pts);

    // Create some data
    for(unsigned int i=0; i < data.size(); i++)
        data[i] = -100.0 + 200.0*(rand()/(1.0 + RAND_MAX));

    for(unsigned int i=0; i < query.size(); i++)
        query[i] = -100.0 + 200.0*(rand()/(1.0 + RAND_MAX));

    gettimeofday(&t1, NULL);
    CUDA_NN_Search(&query[0], query_pts, &data[0], data_pts, &idxs[0], &dist_sq[0]);
    gettimeofday(&t2, NULL);
    double GPU_time = TimeDiff(t1,t2);

    // CPU version
    double CPU_time = SearchCPU(query, data, cpu_idxs, cpu_dist_sq);
    double ANN_time = SearchANN(query, data, cpu_idxs, cpu_dist_sq);

    // Verify correctness
    for(unsigned int i=0; i < cpu_idxs.size(); i++) {
        if(cpu_idxs[i] != idxs[i]) {
            printf("Results do not match :(\n");
            printf("%d - %d\n", cpu_idxs[i], idxs[i]);
            return 1;
        }
    }

    printf("Data dimension: %d\n", CUDA_NN_DIM);
    printf("Total data points: %d\n", data_pts);
    printf("Total query points: %d\n", query_pts);
    printf("\n");
    printf("GPU: %g ms\n", GPU_time);
    printf("CPU brute force: %g ms\n", CPU_time);
    printf("libANN: %g ms\n", ANN_time);
    printf("\n");
    printf("Speed up over CPU brute force from using GPU: %.2fx\n", CPU_time/GPU_time);
    printf("Speed up over libANN using GPU: %.2fx\n", ANN_time/GPU_time);

    return 0;
}

double TimeDiff(timeval t1, timeval t2)
{
    double t;
    t = (t2.tv_sec - t1.tv_sec) * 1000.0;      // sec to ms
    t += (t2.tv_usec - t1.tv_usec) / 1000.0;   // us to ms

    return t;
}

double SearchCPU(const vector <float> &query, const vector <float> &data, vector <int> &idxs, vector <float> dist_sq)
{
    timeval t1, t2;

    int query_pts = query.size() / CUDA_NN_DIM;
    int data_pts = data.size() / CUDA_NN_DIM;

    gettimeofday(&t1, NULL);
    for(unsigned int i=0; i < query_pts; i++) {
        float best_dist = FLT_MAX;
        int best_idx = 0;

        for(unsigned int j=0; j < data_pts; j++) {
            float dist_sq = 0;

            for(int k=0; k < CUDA_NN_DIM; k++) {
                float d = query[i*CUDA_NN_DIM+k] - data[j*CUDA_NN_DIM+k];
                dist_sq += d*d;
            }

            if(dist_sq < best_dist) {
                best_dist = dist_sq;
                best_idx = j;
            }
        }

        idxs[i] = best_idx;
        dist_sq[i] = best_dist;
    }

    gettimeofday(&t2, NULL);

    return TimeDiff(t1,t2);
}

double SearchANN(const vector <float> &query, const vector <float> &data, vector <int> &idxs, vector <float> dist_sq)
{
    int k = 1;

    ANNidxArray nnIdx = new ANNidx[k];
    ANNdistArray dists = new ANNdist[k];
    ANNpoint queryPt = annAllocPt(CUDA_NN_DIM);

    int query_pts = query.size() / CUDA_NN_DIM;
    int data_pts = data.size() / CUDA_NN_DIM;
    timeval t1, t2;

    ANNpointArray dataPts = annAllocPts(data_pts, CUDA_NN_DIM);

    for(unsigned int i=0; i < data_pts; i++) {
        for(int j=0; j < CUDA_NN_DIM; j++ ) {
            dataPts[i][j] = data[i*CUDA_NN_DIM + j];
        }
    }

    gettimeofday(&t1, NULL);
    ANNkd_tree* kdTree = new ANNkd_tree(dataPts, data_pts, CUDA_NN_DIM);

    for(int i=0; i < query_pts; i++) {
        for(int j=0; j < CUDA_NN_DIM; j++) {
            queryPt[j] = query[i*CUDA_NN_DIM+j];
        }

        kdTree->annkSearch(queryPt, 1, nnIdx, dists);

        idxs[i] = nnIdx[0];
        dist_sq[i] = dists[0];
    }
    gettimeofday(&t2, NULL);

	delete [] nnIdx;
	delete [] dists;
	delete kdTree;
	annDeallocPts(dataPts);
	annClose();

    return TimeDiff(t1,t2);
}
